winmessage WM_SUBCLASSED   ;(ptrProcedure, Tag) the window receives this message immediately after subclassing with AddWinProc.

struct TChainElement
  .WinProc dd ?
  .Tag     dd ?
  .count   dd ?
  .next    dd ?
ends


iglobal
  winproperty propWinProcChain, 'WinProcChain'
endg


;initialize InitChainMessages
;begin
;        stdcall CreateWinProperty, propWinProcChain
;        return
;endp
;
;
;finalize FinalChainMessages
;begin
;        stdcall DestroyWinProperty, propWinProcChain
;        return
;endp
;


proc AddWinProc, .hwnd, .winproc, .tag
begin
        push    esi edi

        stdcall SearchChainElement, [.hwnd], [.winproc], [.tag]
        jc      .addit

; The message handler exists, just inc the count
        inc     [esi+TChainElement.count]
        jmp     .exit

; Add new message handler.
.addit:
        stdcall GetMem, sizeof.TChainElement
        mov     esi, eax

        mov     eax, [.winproc]
        mov     ecx, [.tag]
        mov     [esi+TChainElement.WinProc], eax
        mov     [esi+TChainElement.Tag], ecx
        mov     [esi+TChainElement.count], 1

        invoke  GetProp, [.hwnd], [propWinProcChain]
        test    eax, eax
        jz      .newchain

        mov     [esi+TChainElement.next], eax
.save:
        invoke  SetProp, [.hwnd], [propWinProcChain], esi
.exit:
        invoke  SendMessage, [.hwnd], WM_SUBCLASSED, [.winproc], [.tag]
        pop     edi esi
        return

.newchain:
        stdcall GetMem, sizeof.TChainElement
        mov     edi, eax

        invoke  GetWindowLong, [.hwnd], GWL_WNDPROC
        mov     [edi+TChainElement.WinProc], eax
        mov     [esi+TChainElement.next], edi
        invoke  SetWindowLong, [.hwnd], GWL_WNDPROC, ChainWinProc
        jmp     .save
endp




;----------------------------------------------------
; if [.winproc] <> 0 removes the message handle with
; specified [.winproc], [.tag] from the handler chain
; if [.winproc] = 0 removes the first handler from
; the chain.
;----------------------------------------------------
proc RemoveWinProc, .hwnd, .winproc, .tag
begin
        push    edi esi

        stdcall SearchChainElement, [.hwnd], [.winproc], [.tag]
        jc      .exit

        dec     [esi+TChainElement.count]
        jnz     .exit                           ; don't remove, someone needs it.

        test    edi, edi
        jz      .first

        mov     ecx, [esi+TChainElement.next]
        jecxz   .exit   ; can't be removed the original handler.

        mov     [edi+TChainElement.next], ecx

.free:
        stdcall FreeMem, esi
.exit:
        pop     esi edi
        return

.first:
        mov     edi, [esi+TChainElement.next]
        cmp     [edi+TChainElement.next], 0
        je      .removechain

        invoke  SetProp, [.hwnd], [propWinProcChain], edi
        jmp     .free

.removechain:
        invoke  SetWindowLong, [.hwnd], GWL_WNDPROC, [edi+TChainElement.WinProc]
        invoke  RemoveProp, [.hwnd], [propWinProcChain]
        stdcall FreeMem, edi
        jmp     .free
endp



;-----------------------------------------------
; Arguments:
;   .hwnd - handle of the window
;   .winproc - user winproc (if 0 - returns the first in chain)
;   .tag     - user defined value.
; Returns:
;   CF = 0 -> element found.
;   esi - TChainElement
;   edi - previous in chain
;   CF = 1 -> element not found.
;     esi = 0
;     edi = 0 -> the window is not subclassed
;     edi <> 0 -> TChainElement of original winproc
;-----------------------------------------------
proc SearchChainElement, .hwnd, .winproc, .tag
begin
        xor     edi, edi

        invoke  GetProp, [.hwnd], [propWinProcChain]
        mov     esi, eax
        test    eax, eax
        jz      .exit

        cmp     [.winproc], 0
        je      .found

.loop:
        mov     ecx, [esi+TChainElement.WinProc]
        mov     eax, [esi+TChainElement.Tag]

        cmp     ecx, [.winproc]
        jne     .next
        cmp     ecx, [.tag]
        jne     .next

.found:
        clc
        return

.next:
        mov     edi, esi
        mov     esi, [esi+TChainElement.next]
        test    esi, esi
        jnz     .loop

.exit:
        stc
        return
endp


;--------------------------------------------------------------
; This is universal message handler for subclassed windows.
; Saves registers and provides call to the previous handler.
;--------------------------------------------------------------
proc ChainWinProc, .hwnd, .wmsg, .wparam, .lparam
begin
        push    ebx esi edi

        invoke  GetProp, [.hwnd], [propWinProcChain]
        mov     ecx, eax

.loop:
        jecxz   .finish

        push    [ecx+TChainElement.next]      ; in the case current handler removes itself from the chain.
        mov     edi, [ecx+TChainElement.Tag]
        cmp     [ecx+TChainElement.WinProc], 0
        jg      .procedure

        ; handle
        invoke  CallWindowProc, [ecx+TChainElement.WinProc], [.hwnd], [.wmsg], [.wparam], [.lparam]
        jmp     .next

.procedure:
        stdcall [ecx+TChainElement.WinProc], [.hwnd], [.wmsg], [.wparam], [.lparam]

.next:
        pop     ecx
        jc      .loop

.finish:
        cmp     [.wmsg], WM_DESTROY
        jne     .quit

; clear whole chain and remove the property.
        push    eax
        xor     ebx, ebx

.freeloop:
        stdcall RemoveWinProc, [.hwnd], ebx, ebx
        invoke  GetProp, [.hwnd], [propWinProcChain]
        test    eax, eax
        jnz     .freeloop
        pop     eax

.quit:
        pop     edi esi ebx
        return
endp



