;------------------------------------------------------------------------
; Macro for defining names for window properties (SetProp/GetProp)
; that to be used with and without AddAtom function.
; When the name is not added to the atom table, the first
; dword of the structure contains pointer to the string.
; If the macro is added to the application atom table
; (via CreateProperty procedure), the first dword contains
; atom handle. In both cases GetProp and SetProp uses [propname] as
; pointer to the property name.
;------------------------------------------------------------------------
macro winproperty [name, string] {
  forward
    align 4
    local ..txt
    name  dd ..txt
    TEXT ..txt, string
    dw   0
}


; Various property related string procedures
proc CreateWinProperty, .ptrProp
begin
        push    ebx
        mov     ebx, [.ptrProp]
        invoke  IsBadReadPtr, ebx, 4
        test    eax, eax
        jnz     .exit

        invoke  IsBadStringPtr, [ebx], 256
        test    eax, eax
        jnz     .exit

        invoke  AddAtom, [ebx]
        movzx   eax, ax
        mov     [ebx], eax
.exit:
        pop     ebx
        return
endp


proc DestroyWinProperty, .ptrProp
begin
        push    ebx
        mov     ebx, [.ptrProp]
        invoke  IsBadReadPtr, ebx, 4
        test    eax, eax
        jnz     .exit

        invoke  DeleteAtom, [ebx]
.exit:
        pop     ebx
        return
endp