; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "Fresh artistic license."         |
;|_______________________________________________________________________________________|
;
;  Description: Process management library. Win32 part.
;
;  Target OS: Win32
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________


; Terminate current process
; Returns: eax = 0

proc Terminate,.exit_code
begin
        invoke ExitProcess, [.exit_code]
        return
endp



proc GetCmdArguments
.argc dd ?
.array dd ?
begin
        pushad
        invoke  GetCommandLine
        lea     ecx, [.argc]
        invoke  CommandLineToArgv, eax, ecx
        test    eax, eax
        jz      .error

        push    eax
        mov     esi, eax

        stdcall CreateArray, 4
        mov     [.array], eax

.argloop:
        cmp     [.argc], 0
        je      .endarg

        stdcall WideCharToUtf8, [esi]
        push    eax
        stdcall AddArrayItems, [.array], 1
        mov     [.array], edx
        pop     dword [eax]

        add     esi, 4
        dec     [.argc]
        jmp     .argloop

.endarg:
        invoke  LocalFree ; from the stack...
        popad
        mov     eax, [.array]
        clc
        return

.error:
        popad
        stc
        return
endp



; Creates a new execution thread
; Returns: CF=1 if error.
;          CF=0, eax = pid of the new process being created

proc ThreadCreate, .ptr_to_function, .ptr_to_args
begin
        push    ecx edx

        invoke  CreateThread,NULL,0,[.ptr_to_function],[.ptr_to_args],0,0
        test    eax, eax
        jz      .error

        clc
        pop     edx ecx
        return

.error:
        stc
        pop     edx ecx
        return
endp




; MUTEX functions
;_________________________________________________________________________________________

proc MutexCreate, .ptrName, .ptrMutex
begin
        push    eax ecx edx

        invoke  CreateMutex, 0, TRUE, 0
        mov     ecx, [.ptrMutex]
        mov     [ecx], eax
        pop     edx ecx eax
        return
endp




proc WaitForMutex, .ptrMutex, .timeout
begin
        push    eax ecx edx
        mov     ecx, [.ptrMutex]
        invoke  WaitForSingleObject, [ecx], [.timeout]
        cmp     eax, WAIT_TIMEOUT
        clc
        jne     @f
        stc
@@:
        pop     edx ecx eax
        return
endp




proc MutexRelease, .ptrMutex
begin
        pushf
        push    eax ecx edx

        mov     eax, [.ptrMutex]
        invoke  ReleaseMutex, [eax]

        pop     edx ecx eax
        popf
        return
endp



proc MutexDestroy, .ptrMutex
begin
        push    eax ecx edx
        mov     eax, [.ptrMutex]
        pushd   [eax]
        mov     dword [eax], 0
        invoke  CloseHandle ; from the stack
        pop     edx ecx eax
        return
endp

